/** \file
 *
 *  \brief GTK+ 2 user-interface common functions.
 *
 *  \copyright Copyright 2014-2024 Ciaran Anscomb
 *
 *  \licenseblock This file is part of XRoar, a Dragon/Tandy CoCo emulator.
 *
 *  XRoar is free software; you can redistribute it and/or modify it under the
 *  terms of the GNU General Public License as published by the Free Software
 *  Foundation, either version 3 of the License, or (at your option) any later
 *  version.
 *
 *  See COPYING.GPL for redistribution conditions.
 *
 *  \endlicenseblock
 */

#include "top-config.h"

#include <ctype.h>

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wstrict-prototypes"
#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
#include <gtk/gtk.h>
#pragma GCC diagnostic pop
#include <gdk/gdkkeysyms.h>

#include "xalloc.h"

#include "auto_kbd.h"
#include "xroar.h"

#include "gtk2/common.h"

// Eventually, everything should be delegated properly, but for now assure
// there is only ever one instantiation of ui_gtk2 and make it available
// globally.
struct ui_gtk2_interface *global_uigtk2 = NULL;

// Event handlers

// Used within tape/drive control dialogs to eat keypresses but still allow GUI
// controls.

gboolean gtk2_dummy_keypress(GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	(void)widget;
	struct ui_gtk2_interface *uigtk2 = user_data;

	if (gtk_window_activate_key(GTK_WINDOW(uigtk2->top_window), event) == TRUE) {
		return TRUE;
	}

	return FALSE;
}

// Key press/release

gboolean gtk2_handle_key_press(GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	struct ui_gtk2_interface *uigtk2 = user_data;

#ifndef WINDOWS32
	// Hide cursor
	if (!uigtk2->cursor_hidden) {
		GdkWindow *window = gtk_widget_get_window(uigtk2->drawing_area);
		uigtk2->old_cursor = gdk_window_get_cursor(window);
		gdk_window_set_cursor(window, uigtk2->blank_cursor);
		uigtk2->cursor_hidden = 1;
	}
#endif

	// Pass off to keyboard code
	return gtk2_keyboard_handle_key_press(widget, event, user_data);
}

gboolean gtk2_handle_key_release(GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
	// Pass off to keyboard code
	return gtk2_keyboard_handle_key_release(widget, event, user_data);
}

// Pointer motion

gboolean gtk2_handle_motion_notify(GtkWidget *widget, GdkEventMotion *event, gpointer user_data) {
	struct ui_gtk2_interface *uigtk2 = user_data;
	struct vo_interface *vo = uigtk2->public.vo_interface;
	(void)widget;

#ifndef WINDOWS32
	// Unhide cursor
	if (uigtk2->cursor_hidden) {
		GdkWindow *window = gtk_widget_get_window(uigtk2->drawing_area);
		gdk_window_set_cursor(window, uigtk2->old_cursor);
		uigtk2->cursor_hidden = 0;
	}
#endif

	// Update position data (for mouse mapped joystick)
	vo->mouse.axis[0] = event->x;
	vo->mouse.axis[1] = event->y;

	return FALSE;
}

// Button press/release

static void clipboard_text_received(GtkClipboard *clipboard, const gchar *text, gpointer data) {
	(void)clipboard;
	(void)data;
	if (!text)
		return;
	char *ntext = xstrdup(text);
	if (!ntext)
		return;
	guint state = (uintptr_t)data;
	_Bool uc = state & GDK_SHIFT_MASK;
	for (char *p = ntext; *p; p++) {
		if (*p == '\n')
			*p = '\r';
		if (uc)
			*p = toupper(*p);
	}
	ak_parse_type_string(xroar.auto_kbd, ntext);
	free(ntext);
}

gboolean gtk2_handle_button_press(GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	struct ui_gtk2_interface *uigtk2 = user_data;
	struct vo_interface *vo = uigtk2->public.vo_interface;
	(void)widget;

	if (event->button == 2) {
		GdkDisplay *d = gtk_widget_get_display(uigtk2->top_window);
		GtkClipboard *cb = gtk_clipboard_get_for_display(d, GDK_SELECTION_PRIMARY);
		gtk_clipboard_request_text(cb, clipboard_text_received, (gpointer)(uintptr_t)event->state);
		return FALSE;
	}

	// Update button data (for mouse mapped joystick)
	if (event->button >= 1 && event->button <= 3) {
		vo->mouse.button[event->button-1] = 1;
	}

	return FALSE;
}

gboolean gtk2_handle_button_release(GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
	struct ui_gtk2_interface *uigtk2 = user_data;
	struct vo_interface *vo = uigtk2->public.vo_interface;
	(void)widget;

	// Update button data (for mouse mapped joystick)
	if (event->button >= 1 && event->button <= 3) {
		vo->mouse.button[event->button-1] = 0;
	}

	return FALSE;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// UI builder helpers

FUNC_ATTR_NORETURN static void do_g_abort(const gchar *format, GError *error) {
	(void)format;
	if (error) {
		g_message("gtk_builder_add_from_resource() failed: %s", error->message);
		g_error_free(error);
	}
	g_abort();
}

void uigtk2_add_from_resource(struct ui_gtk2_interface *uigtk2, const gchar *path) {
	GError *error = NULL;
	GBytes *resource = g_resources_lookup_data(path, 0, &error);
	if (!resource) {
		do_g_abort("g_resources_lookup_data() failed: %s", error);
	}

	gsize xml_size;
	const gchar *xml = g_bytes_get_data(resource, &xml_size);

	if (gtk_builder_add_from_string(uigtk2->builder, xml, xml_size, &error) == 0) {
		do_g_abort("gtk_builder_add_from_string() failed: %s", error);
	}

	g_bytes_unref(resource);
}

void do_uigtk2_signal_connect(struct ui_gtk2_interface *uigtk2, const gchar *o_name,
			      const gchar *detailed_signal,
			      GCallback c_handler,
			      gpointer data) {
	GObject *o = gtk_builder_get_object(uigtk2->builder, o_name);
	g_signal_connect(o, detailed_signal, c_handler, data);
}

// Notify-only menu manager update helpers.
//
// Blocks callback so that no further action is taken.

void uigtk2_notify_radio_action_set_current_value(struct ui_gtk2_interface *uigtk2,
						  const gchar *path, gint v, gpointer func) {
	GtkRadioAction *ra = GTK_RADIO_ACTION(gtk_ui_manager_get_action(uigtk2->menu_manager, path));
	g_signal_handlers_block_by_func(ra, G_CALLBACK(func), uigtk2);
	gtk_radio_action_set_current_value(ra, v);
	g_signal_handlers_unblock_by_func(ra, G_CALLBACK(func), uigtk2);
}

void uigtk2_notify_radio_menu_set_current_value(struct uigtk2_radio_menu *rm, gint v) {
	if (!rm)
		return;
	GList *list = gtk_action_group_list_actions(rm->action_group);
	if (!list)
		return;
	GtkRadioAction *ra = GTK_RADIO_ACTION(list->data);
	g_list_free(list);
	g_signal_handlers_block_by_func(ra, rm->callback, rm->uigtk2);
	gtk_radio_action_set_current_value(ra, v);
	g_signal_handlers_unblock_by_func(ra, rm->callback, rm->uigtk2);
}

void uigtk2_notify_toggle_action_set_active(struct ui_gtk2_interface *uigtk2,
					    const gchar *path, gboolean v, gpointer func) {
	GtkToggleAction *ta = GTK_TOGGLE_ACTION(gtk_ui_manager_get_action(uigtk2->menu_manager, path));
	g_signal_handlers_block_by_func(ta, G_CALLBACK(func), uigtk2);
	gtk_toggle_action_set_active(ta, v);
	g_signal_handlers_unblock_by_func(ta, G_CALLBACK(func), uigtk2);
}

// Notify-only UI update helpers.
//
// Blocks callback so that no further action is taken.

void uigtk2_notify_spin_button_set_value(struct ui_gtk2_interface *uigtk2,
					 const gchar *sb_name, gdouble value, gpointer func) {
	GtkSpinButton *sb = GTK_SPIN_BUTTON(gtk_builder_get_object(uigtk2->builder, sb_name));
	g_signal_handlers_block_by_func(sb, G_CALLBACK(func), uigtk2);
	gtk_spin_button_set_value(sb, value);
	g_signal_handlers_unblock_by_func(sb, G_CALLBACK(func), uigtk2);
}

void uigtk2_notify_toggle_button_set_active(struct ui_gtk2_interface *uigtk2,
					    const gchar *tb_name,
					    gboolean v, gpointer func) {
	GtkToggleButton *tb = GTK_TOGGLE_BUTTON(gtk_builder_get_object(uigtk2->builder, tb_name));
	g_signal_handlers_block_by_func(tb, G_CALLBACK(func), uigtk2);
	gtk_toggle_button_set_active(tb, v);
	g_signal_handlers_unblock_by_func(tb, G_CALLBACK(func), uigtk2);
}

// Menu manager helpers

gboolean uigtk2_toggle_action_get_active(struct ui_gtk2_interface *uigtk2, const gchar *path) {
	GtkToggleAction *ta = GTK_TOGGLE_ACTION(gtk_ui_manager_get_action(uigtk2->menu_manager, path));
	return gtk_toggle_action_get_active(ta);
}

void uigtk2_toggle_action_set_active(struct ui_gtk2_interface *uigtk2, const gchar *path,
				     gboolean v) {
	GtkToggleAction *ta = GTK_TOGGLE_ACTION(gtk_ui_manager_get_action(uigtk2->menu_manager, path));
	gtk_toggle_action_set_active(ta, v);
}

// UI helpers

void uigtk2_adjustment_set_lower(struct ui_gtk2_interface *uigtk2, const gchar *a_name,
				 gdouble lower) {
	GtkAdjustment *a = GTK_ADJUSTMENT(gtk_builder_get_object(uigtk2->builder, a_name));
	gtk_adjustment_set_lower(a, lower);
}

void uigtk2_adjustment_set_upper(struct ui_gtk2_interface *uigtk2, const gchar *a_name,
				 gdouble upper) {
	GtkAdjustment *a = GTK_ADJUSTMENT(gtk_builder_get_object(uigtk2->builder, a_name));
	gtk_adjustment_set_upper(a, upper);
}

void uigtk2_adjustment_set_value(struct ui_gtk2_interface *uigtk2, const gchar *a_name,
				 gdouble value) {
	GtkAdjustment *a = GTK_ADJUSTMENT(gtk_builder_get_object(uigtk2->builder, a_name));
	gtk_adjustment_set_value(a, value);
}

void uigtk2_combo_box_set_active(struct ui_gtk2_interface *uigtk2, const gchar *cbt_name,
				 gint index_) {
	GtkComboBoxText *cbt = GTK_COMBO_BOX_TEXT(gtk_builder_get_object(uigtk2->builder, cbt_name));
	gtk_combo_box_set_active(GTK_COMBO_BOX(cbt), index_);
}

void uigtk2_label_set_text(struct ui_gtk2_interface *uigtk2, const gchar *l_name,
			   const gchar *str) {
	GtkLabel *l = GTK_LABEL(gtk_builder_get_object(uigtk2->builder, l_name));
	gtk_label_set_text(l, str);
}

void uigtk2_toggle_button_set_active(struct ui_gtk2_interface *uigtk2, const gchar *tb_name,
				     gboolean v) {
	GtkToggleButton *tb = GTK_TOGGLE_BUTTON(gtk_builder_get_object(uigtk2->builder, tb_name));
	gtk_toggle_button_set_active(tb, v);
}

void uigtk2_widget_hide(struct ui_gtk2_interface *uigtk2, const gchar *w_name) {
	GtkWidget *w = GTK_WIDGET(gtk_builder_get_object(uigtk2->builder, w_name));
	gtk_widget_hide(w);
}

void uigtk2_widget_set_sensitive(struct ui_gtk2_interface *uigtk2, const gchar *w_name,
				 gboolean sensitive) {
	GtkWidget *w = GTK_WIDGET(gtk_builder_get_object(uigtk2->builder, w_name));
	gtk_widget_set_sensitive(w, sensitive);
}

void uigtk2_widget_show(struct ui_gtk2_interface *uigtk2, const gchar *w_name) {
	GtkWidget *w = GTK_WIDGET(gtk_builder_get_object(uigtk2->builder, w_name));
	gtk_widget_show(w);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// Create all the required bits & bobs for keeping a dynamically created radio
// menu updated.

struct uigtk2_radio_menu *uigtk2_radio_menu_new(struct ui_gtk2_interface *uigtk2,
						const char *path, GCallback callback) {
	struct uigtk2_radio_menu *rm = g_malloc0(sizeof(*rm));
	static unsigned id = 0;
	rm->uigtk2 = uigtk2;
	rm->path = g_strdup(path);
	rm->action_group_name = g_strdup_printf("rm%u", id++);
	rm->action_group = gtk_action_group_new(rm->action_group_name);
	gtk_ui_manager_insert_action_group(uigtk2->menu_manager, rm->action_group, -1);
	rm->merge_id = gtk_ui_manager_new_merge_id(uigtk2->menu_manager);
	rm->callback = callback;
	return rm;
}

void uigtk2_radio_menu_free(struct uigtk2_radio_menu *rm) {
	struct ui_gtk2_interface *uigtk2 = rm->uigtk2;
	gtk_ui_manager_remove_action_group(uigtk2->menu_manager, rm->action_group);
	g_object_unref(rm->action_group);
	g_free(rm->action_group_name);
	g_free(rm->path);
	g_free(rm);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// [Re-]build a menu from an xconfig_enum

void uigtk2_update_radio_menu_from_enum(struct uigtk2_radio_menu *rm,
					struct xconfig_enum *xc_enum,
					const char *name_fmt, const char *label_fmt,
					int selected) {
	if (!rm || !xc_enum)
		return;
	struct ui_gtk2_interface *uigtk2 = rm->uigtk2;
	if (!name_fmt)
		name_fmt = "%s";
	if (!label_fmt)
		label_fmt = "%s";

	// Remove old entries
	uigtk2_free_action_group(rm->action_group);
	gtk_ui_manager_remove_ui(uigtk2->menu_manager, rm->merge_id);

	// Count entries
	int num_entries = 0;
	int enum_index = 0;
	for (enum_index = 0; xc_enum[enum_index].name; enum_index++) {
		if (!xc_enum[enum_index].description)
			continue;
		num_entries++;
	}

	// Add entries in reverse order
	GtkRadioActionEntry *entries = g_malloc0(num_entries * sizeof(*entries));
	gchar **names = g_malloc0(num_entries * sizeof(gchar *));
	gchar **labels = g_malloc0(num_entries * sizeof(gchar *));
	for (int i = 0; i < num_entries && enum_index > 0; ) {
		--enum_index;
		if (!xc_enum[enum_index].description)
			continue;
		names[i] = g_strdup_printf(name_fmt, xc_enum[enum_index].name);
		labels[i] = g_strdup_printf(label_fmt, xc_enum[enum_index].description);
		entries[i].name = names[i];
		entries[i].label = xc_enum[enum_index].description;
		entries[i].value = xc_enum[enum_index].value;
		gtk_ui_manager_add_ui(uigtk2->menu_manager, rm->merge_id, rm->path, entries[i].name, entries[i].name, GTK_UI_MANAGER_MENUITEM, TRUE);
		++i;
	}
	gtk_action_group_add_radio_actions(rm->action_group, entries, num_entries, selected, rm->callback, uigtk2);

	// Free everything
	for (int i = 0; i < num_entries; i++) {
		g_free(names[i]);
		g_free(labels[i]);
	}
	g_free(names);
	g_free(labels);
	g_free(entries);
}

static void remove_action_from_group(gpointer data, gpointer user_data) {
	GtkAction *action = data;
	GtkActionGroup *action_group = user_data;
	gtk_action_group_remove_action(action_group, action);
}

void uigtk2_free_action_group(GtkActionGroup *action_group) {
	GList *list = gtk_action_group_list_actions(action_group);
	g_list_foreach(list, remove_action_from_group, action_group);
	g_list_free(list);
}
